from .claudiotrack import ClAudioTrack
from .clmiditrack import ClMidiTrack
from .consts import *
from .dltrack import DlTrack
from .paramtracker import ParamTracker
from .cltrack import ClTrack
from .clovrtrack import ClOVRTrack
from collections import defaultdict


class TrackHandler:


    """ Class handling looper & clip tracks """
    def __init__(self, parent, song, tracks, state, actions):
        self.__parent = parent
        self.song = song
        self.actions = actions
        self.tracks = tracks
        self.trackStore = []
        self.track_nums = []
        self.param_store = []
        self.state = state
        self.overdub_tracks = {}
        self.looper_state_listeners = []

    def disconnect(self):
        self.__parent.disconnect()

    # Detects tracks with 'looper' in the title and listens for param changes
    def scan_tracks(self, isInit = False):
        self.send_message("scanning for DataLooper tracks")

        # clear CL# identified tracks
        self.clear_tracks()
        self.track_nums = []
        self.iterate_tracks(self.song.tracks, isInit)
        self.iterate_tracks(self.song.return_tracks, isInit)
        self.send_sysex(0x00, 0x01)

        for tracks in self.tracks.values():
            led_track = self.get_led_track(tracks)
            if len(tracks) > 1 and led_track:
                for track in tracks:
                    if track is not led_track:
                        self.send_message(track)
                        track.disable_led()

    def iterate_tracks(self, tracks, isInit):
        """ 
        ALTERNATE DELIMITERS
        OVR --Indicates this track should be used as an overdub track for CLAudioTrack
        LED --Indicates that this track should be used as the master for LED Control, in the case that there are multiple tracks with the same #
        #G1 --Indicates part of group 1 (not necessarily connected to a DL track, could be a folder track)
        #M1 --Indicates master of group 1

        Master track could control multiple groups, defaults to master #, but sysex command could change group connectivity

        If master 1 starts playback, group 1 loopers all mute

        #G1 --- on all loopers in a group...group track gets affected.
        
        SMART CLEAR
        Clear any loopers in muted blocks that are unmuted and playing
        """

            
        # iterate through all tracks
        for idx, track in enumerate(tracks):
            # check for tracks with naming convention
            for key in TRACK_KEYS:
                # checks if key exists in name
                string_pos = track.name.find(key)
                if string_pos != -1:
                    # Checks for double digits
                    if len(track.name) >= string_pos + 5 and track.name[string_pos + 4: string_pos + 5].isdigit():
                        track_num = int(track.name[string_pos + 3: string_pos + 5]) - 1
                    else:
                        try:
                            track_num = int(track.name[string_pos + 3: string_pos + 4]) - 1
                        except ValueError:
                            self.send_message("Track needs a number after the #")
                            continue
                    self.track_nums.append(track_num)
                    self.append_tracks(track, track_num, key)
            # if isInit:
                # self.param_store.append(ParamTracker(track, self, self.song, idx))
            # adds name change listener to all tracks
            if not track.name_has_listener(self.__parent.on_track_name_changed):
                track.add_name_listener(self.__parent.on_track_name_changed)

        self.send_message("scanning overdub tracks")
        for idx, overdub_track in enumerate(self.overdub_tracks):
            self.send_message("index is:" + str(idx))
            for track in self.tracks[str(overdub_track)]:
                if isinstance(track, ClAudioTrack):
                    self.send_message("linking overdub track on track number" + str(idx))
                    track.link_overdub(self.overdub_tracks.get(overdub_track))

    def get_led_track(self, tracks):
        for track in tracks:
            if "LED" in track.track.name:
                return track
        return False

    # def clear_unused_tracks(self, track_nums):
    #     # Sends clear to tracks on pedal that aren't linked. IE, if there's CL#1 & CL#2, track 3 will get a clear state
    #     i = 0
    #     while i < NUM_TRACKS:
    #         if i not in track_nums:
    #             self.__parent.send_sysex(DLCommands.CHANGE_STATE,i, CLEAR_STATE)
    #         i += 1

    def clear_tracks(self):
        msg = "Removing Tracks: "
        for trackNums in self.tracks.values():
            for track in trackNums:
                track.remove_track()
                if track.track in self.song.tracks:
                    msg += track.track.name + " "

        self.tracks.clear()
        self.send_message(msg)

    def add_looper_state_listener(self, listener):
        self.looper_state_listeners.append(listener)

    def send_looper_state(self, track_num, state):
        for listener in self.looper_state_listeners:
            listener.on_looper_state_updated(track_num, state)

    def append_tracks(self, track, trackNum, track_key):
        msg = "Adding tracks: "
        track_object = None
        if track_key == DATALOOPER_KEY:
            # adds a listener to tracks detected as DataLoopers to rescan for looper when a devices is added
            if not track.devices_has_listener(self.scan_tracks):
                track.add_devices_listener(self.scan_tracks)
            # stops adding track if there's other DL# with same number, to avoid duplicated note messages
            for ex_track in self.tracks[str(trackNum)]:
                self.send_message("track on " + str(trackNum))
                if isinstance(ex_track, DlTrack):
                    self.send_message("duplicates found on track number" + str(trackNum))
                    ex_track.set_duplicate(True)
            # checks for devices
            if track.devices:
                for device in track.devices:
                    if device.can_have_chains and track_object is None:
                        for chain in device.chains:
                            if chain.devices:
                                for chain_device in chain.devices:
                                    msg += track.name
                                    if chain_device.class_name == "Looper":
                                        track_object = DlTrack(self, track, chain_device, trackNum, self.song, self.state, self.actions)
                                        break
                    elif device.class_name == "Looper":
                        msg += track.name
                        track_object = DlTrack(self, track, device, trackNum, self.song, self.state, self.actions)
                        break
                    elif track_object != None:
                        break
        elif track_key == CLIPLOOPER_KEY:
            msg += track.name
            if track.has_midi_input:
                track_object = ClMidiTrack(self, track, trackNum, self.song, self.state, self.actions)
            elif track.has_audio_input:
                if track.name.find("OVR") != -1:
                    self.send_message("adding overdub track on track number" + str(trackNum))
                    self.overdub_tracks[str(trackNum)] = track
                else:
                    track_object = ClAudioTrack(self, track, trackNum, self.song, self.state, self.actions)
        if track_object != None:   
            self.scan_secondary_delimiters(track_object)
            self.tracks[str(trackNum)].append(track_object)
        self.send_message(msg)
    
    def check_for_looper(self,device,track,trackNum):
        if device.class_name == "Looper":
            return DlTrack(self, track, device, trackNum, self.song, self.state, self.actions)
    
    def scan_secondary_delimiters(self, track_object):
        self.send_message("Scanning secondary delimiters")
        name = track_object.track.name
        for key in ALT_KEYS:
                # checks if key exists in name
            string_pos = name.find(key)
            string_pos + len(key) + 1

            if string_pos != -1:
                method = self.get_delimiter_method(key)
                self.send_message("Found key" + str(key) + " Calling delimiter method " + str(method) + " name:" + str(name))
                num = name[string_pos + len(key)]
                if num.isdigit():
                    getattr(track_object, method)([int(num)])
                else:
                    getattr(track_object, method)()
    
    def get_delimiter_method(self,key):
        return key_map.get(key, "Invalid Action")

    def send_midi(self, midi):
        self.__parent.send_midi(midi)

    def send_message(self, message):
        self.__parent.send_message(message)

    def send_sysex(self, *data):
        self.__parent.send_sysex(*data)

    def session_record(self, overdubbing, trackNum):
        if overdubbing and self.song.session_record == 0:
            for track in self.song.tracks:
                if "CL#"+str(trackNum+1) not in track.name and track.can_be_armed:
                    self.trackStore.append(TempTrack(track.name, track.arm, track.current_monitoring_state))
                    if track.arm == 1:
                        track.arm = 0
                        if track.current_monitoring_state == 1 and track.playing_slot_index == -1:
                            track.current_monitoring_state = 0
        elif not overdubbing and self.song.session_record == 1:
            for track in self.song.tracks:
                if "CL#"+str(trackNum+1) not in track.name and track.can_be_armed:
                    match = next((trackS for trackS in self.trackStore if track.name == trackS.name), None)
                    if match is not None:
                        track.current_monitoring_state = match.current_monitoring_state
                        track.arm = match.arm

    def find_last_slot(self):
        index = []
        for tracks in self.tracks.values():
            for cl_track in tracks:
                if isinstance(cl_track, ClTrack):
                    index.append(cl_track.find_last_slot())
        return max(index)

class TempTrack(object):
    def __init__(self, name, arm, current_monitoring_state):
        self.name = name
        self.arm = arm
        self.current_monitoring_state = current_monitoring_state
